class Error(Exception):
    """Base exception"""


class Warning(Exception):
    pass


class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass


class InternalError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class DataError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    pass


class UnexpectedESInitError(Error):
    """ Should never happen, when a cursor is requested
    without an ElasticSearch object being initialized"""


class UnexpectedRequestResponse(Error):
    """ When perform request returns False, only when HTTP method HEAD
    and status code 404 """



import logging
from types import ModuleType
from typing import Any, Dict, List, Optional

from es import basesqlalchemy
import es.elastic
from sqlalchemy.engine import Connection
import time 

logger = logging.getLogger(__name__)


class ESCompiler(basesqlalchemy.BaseESCompiler):
    pass


class ESTypeCompiler(basesqlalchemy.BaseESTypeCompiler):
    pass


class ESDialect(basesqlalchemy.BaseESDialect):

    name = "elasticsearch"
    scheme = "http"
    driver = "rest"
    statement_compiler = ESCompiler
    type_compiler = ESTypeCompiler

    @classmethod
    def dbapi(cls) -> ModuleType:
        return es.elastic

    def get_table_names(
        self, connection: Connection, schema: Optional[str] = None, **kwargs: Any
    ) -> List[str]:
        query = "SHOW VALID_TABLES"
        result = connection.execute(query)
        open('/mnt/c/Users/PC/Desktop/TMP/sqlalch/log.log','a').write(str(time.time()//1) + ' get table names\n')
        # return a list of table names exclude hidden and empty indexes
        return [table.name for table in result if table.name[0] != "."]

    def get_view_names(
        self, connection: Connection, schema: Optional[str] = None, **kwargs: Any
    ) -> List[str]:
        query = "SHOW VALID_VIEWS"
        result = connection.execute(query)
        # return a list of view names (ES aliases) exclude hidden and empty indexes
        return [table.name for table in result if table.name[0] != "."]

    def get_columns(
        self,
        connection: Connection,
        table_name: str,
        schema: Optional[str] = None,
        **kwargs: Any,
    ) -> List[Dict[str, Any]]:
        query = f'SHOW COLUMNS FROM "{table_name}"'
        # Custom SQL
        array_columns_ = connection.execute(
            f"SHOW ARRAY_COLUMNS FROM {table_name}"
        ).fetchall()
        # convert cursor rows: List[Tuple[str]] to List[str]
        if not array_columns_:
            array_columns = []
        else:
            array_columns = [col_name.name for col_name in array_columns_]

        all_columns = connection.execute(query)
        return [
            {
                "name": row.column,
                "type": basesqlalchemy.get_type(row.mapping),
                "nullable": True,
                "default": None,
            }
            for row in all_columns
            if row.mapping not in self._not_supported_column_types
            and row.column not in array_columns
        ]


ESHTTPDialect = ESDialect


class ESHTTPSDialect(ESDialect):

    scheme = "https"
    default_paramstyle = "pyformat"



import re
from typing import Any, Dict, List, Optional, Tuple
import time

from elasticsearch import Elasticsearch, exceptions as es_exceptions
from es import exceptions
from es.baseapi import (
    apply_parameters,
    BaseConnection,
    BaseCursor,
    check_closed,
    CursorDescriptionRow,
    get_description_from_columns,
    Type,
)
from packaging import version


def connect(
    host: str = "localhost",
    port: int = 9200,
    path: str = "",
    scheme: str = "http",
    user: Optional[str] = None,
    password: Optional[str] = None,
    context: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> BaseConnection:
    """
    Constructor for creating a connection to the database.

        >>> conn = connect('localhost', 9200)
        >>> curs = conn.cursor()

    """
    context = context or {}
    return Connection(host, port, path, scheme, user, password, context, **kwargs)


class Connection(BaseConnection):

    """Connection to an ES Cluster"""

    def __init__(
        self,
        host: str = "localhost",
        port: int = 9200,
        path: str = "",
        scheme: str = "http",
        user: Optional[str] = None,
        password: Optional[str] = None,
        context: Optional[Dict[Any, Any]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            host=host,
            port=port,
            path=path,
            scheme=scheme,
            user=user,
            password=password,
            context=context,
            **kwargs,
        )
        if user and password:
            self.es = Elasticsearch(self.url, http_auth=(user, password), **self.kwargs)
        else:
            self.es = Elasticsearch(self.url, **self.kwargs)

    @check_closed
    def cursor(self) -> BaseCursor:
        """Return a new Cursor Object using the connection."""
        if self.es:
            open('/mnt/c/Users/PC/Desktop/TMP/sqlalch/log.log','a').write(str(time.time()//1) + ' create cursor '+ str(self.url) +'\n')
            cursor = Cursor(self.url, self.es, **self.kwargs)
            self.cursors.append(cursor)
            return cursor
        raise exceptions.UnexpectedESInitError()


class Cursor(BaseCursor):

    """Connection cursor."""

    custom_sql_to_method = {
        "show valid_tables": "get_valid_table_names",
        "show valid_views": "get_valid_view_names",
    }

    def __init__(self, url: str, es: Elasticsearch, **kwargs: Any) -> None:
        super().__init__(url, es, **kwargs)
        self.sql_path = kwargs.get("sql_path") or "_sql"

    def get_valid_table_view_names(self, type_filter: str) -> "Cursor":
        """
        Custom for "SHOW VALID_TABLES" excludes empty indices from the response
        Mixes `SHOW TABLES` with direct index access info to exclude indexes
        that have no rows so no columns (unless templated). SQLAlchemy will
        not support reflection of tables with no columns

        https://github.com/preset-io/elasticsearch-dbapi/issues/38

        :param: type_filter will filter SHOW_TABLES result by BASE_TABLE or VIEW
        """
        results = self.execute("SHOW TABLES")
        response = self.es.cat.indices(format="json")

        _results = []
        for result in results:
            is_empty = False
            for item in response:
                # First column is TABLE_NAME
                if item["index"] == result[0]:
                    if int(item["docs.count"]) == 0:
                        is_empty = True
                        break
            if not is_empty and result[1] == type_filter:
                _results.append(result)
        self._results = _results
        return self

    def get_valid_table_names(self) -> "Cursor":
        # Get the ES cluster version. Since 7.10 the table column name changed #52
        cluster_info = self.es.info()
        cluster_version = version.parse(cluster_info["version"]["number"])
        if cluster_version >= version.parse("7.10.0"):
            return self.get_valid_table_view_names("TABLE")
        return self.get_valid_table_view_names("BASE TABLE")

    def get_valid_view_names(self) -> "Cursor":
        return self.get_valid_table_view_names("VIEW")

    @check_closed
    def execute(
        self, operation: str, parameters: Optional[Dict[str, Any]] = None
    ) -> "BaseCursor":
        open('/mnt/c/Users/PC/Desktop/TMP/sqlalch/log.log','a').write(str(time.time()//1) + ' execute called for '+ str(operation) +'\n')
        cursor = self.custom_sql_to_method_dispatcher(operation)
        if cursor:
            open('/mnt/c/Users/PC/Desktop/TMP/sqlalch/log.log','a').write(str(time.time()//1) + ' execute returned from custom sql dispatcher'+ str(cursor) +'\n')
            return cursor
        open('/mnt/c/Users/PC/Desktop/TMP/sqlalch/log.log','a').write(str(time.time()//1) + ' execute returned null\n')

        re_table_name = re.match("SHOW ARRAY_COLUMNS FROM (.*)", operation)
        if re_table_name:
            return self.get_array_type_columns(re_table_name[1])

        open('/mnt/c/Users/PC/Desktop/TMP/sqlalch/log.log','a').write(str(time.time()//1) + ' creating query with ' + str(operation) + ' : ' + str(parameters) + '\n')
        query = apply_parameters(operation, parameters)
        open('/mnt/c/Users/PC/Desktop/TMP/sqlalch/log.log','a').write(str(time.time()//1) + ' created query |' + str(query) + '|\n')
        results = self.elastic_query(query)
        # We need a list of tuples
        rows = [tuple(row) for row in results.get("rows", [])]
        columns = results.get("columns")
        if not columns:
            raise exceptions.DataError(
                "Missing columns field, maybe it's an opendistro sql ep"
            )
        self._results = rows
        open('/mnt/c/Users/PC/Desktop/TMP/sqlalch/log.log','a').write(str(time.time()//1) + ' results from query |' + str(rows) + '|\n')
        # self._results = [('ASDF',)]
        self.description = get_description_from_columns(columns)
        open('/mnt/c/Users/PC/Desktop/TMP/sqlalch/log.log','a').write(str(time.time()//1) + ' returning |' + str(self) + '|\n')
        return self

    def get_array_type_columns(self, table_name: str) -> "Cursor":
        """
        Queries the index (table) for just one record
        and return a list of array type columns.
        This is useful since arrays are not supported by ES SQL
        """
        array_columns: List[Tuple[Any, ...]] = []
        try:
            response = self.es.search(index=table_name, size=1)
        except es_exceptions.ConnectionError as e:
            raise exceptions.OperationalError(
                f"Error connecting to {self.url}: {e.info}"
            )
        except es_exceptions.NotFoundError as e:
            raise exceptions.ProgrammingError(f"Error ({e.error}): {e.info}")
        try:
            if response["hits"]["total"]["value"] == 0:
                source = {}
            else:
                source = response["hits"]["hits"][0]["_source"]
        except KeyError as e:
            raise exceptions.DataError(
                f"Error inferring array type columns {self.url}: {e}"
            )
        for col_name, value in source.items():
            # If it's a list (ES Array add to cursor)
            if isinstance(value, list):
                if len(value) > 0:
                    # If it's an array of objects add all keys
                    if isinstance(value[0], dict):
                        for in_col_name in value[0]:
                            array_columns.append((f"{col_name}.{in_col_name}",))
                            array_columns.append((f"{col_name}.{in_col_name}.keyword",))
                        continue
                array_columns.append((col_name,))
                array_columns.append((f"{col_name}.keyword",))
        if not array_columns:
            array_columns = []
        self.description = [
            CursorDescriptionRow("name", Type.STRING, None, None, None, None, None)
        ]
        self._results = array_columns
        return self



import os
import unittest

from .fixtures.fixtures import (
    create_alias,
    delete_alias,
    delete_index,
    import_data1,
    import_empty_index,
    import_flights,
)

BASE_URL = "http://localhost:9200"


class TestData(unittest.TestCase):
    def setUp(self):
        self.base_url = os.environ.get("ES_URI", BASE_URL)

    def test_1_data_flights(self):
        delete_index(self.base_url, "flights")
        import_flights(self.base_url)

    def test_2_data_data1(self):
        delete_index(self.base_url, "data1")
        import_data1(self.base_url)

    def test_3_data_empty_index(self):
        delete_index(self.base_url, "empty_index")
        import_empty_index(self.base_url)

    def test_4_alias_to_data1(self):
        alias_name = "alias_to_data1"
        delete_alias(self.base_url, alias_name, "data1")
        create_alias(self.base_url, alias_name, "data1")



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
from typing import Any, Dict, List, Optional, Tuple

from elasticsearch import Elasticsearch, RequestsHttpConnection
from elasticsearch.exceptions import ConnectionError
from es import exceptions
from es.baseapi import (
    apply_parameters,
    BaseConnection,
    BaseCursor,
    check_closed,
    get_description_from_columns,
)
from es.const import DEFAULT_SCHEMA


def connect(
    host: str = "localhost",
    port: int = 443,
    path: str = "",
    scheme: str = "https",
    user: Optional[str] = None,
    password: Optional[str] = None,
    context: Optional[Dict[Any, Any]] = None,
    **kwargs: Any,
) -> BaseConnection:
    """
    Constructor for creating a connection to the database.

        >>> conn = connect('localhost', 9200)
        >>> curs = conn.cursor()

    """
    context = context or {}
    return Connection(host, port, path, scheme, user, password, context, **kwargs)


class Connection(BaseConnection):

    """Connection to an ES Cluster"""

    def __init__(
        self,
        host: str = "localhost",
        port: int = 443,
        path: str = "",
        scheme: str = "https",
        user: Optional[str] = None,
        password: Optional[str] = None,
        context: Optional[Dict[Any, Any]] = None,
        **kwargs: Any,
    ):
        super().__init__(
            host=host,
            port=port,
            path=path,
            scheme=scheme,
            user=user,
            password=password,
            context=context,
            **kwargs,
        )
        if user and password and "aws_keys" not in kwargs:
            self.es = Elasticsearch(self.url, http_auth=(user, password), **self.kwargs)
        # AWS configured credentials on the connection string
        elif user and password and "aws_keys" in kwargs and "aws_region" in kwargs:
            aws_auth = self._aws_auth(user, password, kwargs["aws_region"])
            kwargs.pop("aws_keys")
            kwargs.pop("aws_region")

            self.es = Elasticsearch(
                self.url,
                http_auth=aws_auth,
                connection_class=RequestsHttpConnection,
                **kwargs,
            )
        # aws_profile=<region>
        elif "aws_profile" in kwargs:
            aws_auth = self._aws_auth_profile(kwargs["aws_profile"])
            self.es = Elasticsearch(
                self.url,
                http_auth=aws_auth,
                connection_class=RequestsHttpConnection,
                **kwargs,
            )
        else:
            self.es = Elasticsearch(self.url, **self.kwargs)

    @staticmethod
    def _aws_auth_profile(region: str) -> Any:
        from requests_aws4auth import AWS4Auth
        import boto3

        service = "es"
        credentials = boto3.Session().get_credentials()
        return AWS4Auth(
            credentials.access_key,
            credentials.secret_key,
            region,
            service,
            session_token=credentials.token,
        )

    @staticmethod
    def _aws_auth(aws_access_key: str, aws_secret_key: str, region: str) -> Any:
        from requests_aws4auth import AWS4Auth

        return AWS4Auth(aws_access_key, aws_secret_key, region, "es")

    @check_closed
    def cursor(self) -> "Cursor":
        """Return a new Cursor Object using the connection."""
        if self.es:
            cursor = Cursor(self.url, self.es, **self.kwargs)
            self.cursors.append(cursor)
            return cursor
        raise exceptions.UnexpectedESInitError()


class Cursor(BaseCursor):

    custom_sql_to_method = {
        "show valid_tables": "get_valid_table_names",
        "show valid_views": "get_valid_view_names",
        "select 1": "get_valid_select_one",
    }

    def __init__(self, url: str, es: Elasticsearch, **kwargs: Any) -> None:
        super().__init__(url, es, **kwargs)
        self.sql_path = kwargs.get("sql_path") or "_opendistro/_sql"
        # Opendistro SQL v2 flag
        self.v2 = kwargs.get("v2", False)
        if self.v2:
            self.fetch_size = None

    def get_valid_table_names(self) -> "Cursor":
        """
        Custom for "SHOW VALID_TABLES" excludes empty indices from the response
        Mixes `SHOW TABLES LIKE` with direct index access info to exclude indexes
        that have no rows so no columns (unless templated). SQLAlchemy will
        not support reflection of tables with no columns

        https://github.com/preset-io/elasticsearch-dbapi/issues/38
        """
        results = self.execute("SHOW TABLES LIKE %")
        response = self.es.cat.indices(format="json")

        _results = []
        for result in results:
            is_empty = False
            for item in response:
                # Third column is TABLE_NAME
                if item["index"] == result[2]:
                    if int(item["docs.count"]) == 0:
                        is_empty = True
                        break
            if not is_empty:
                _results.append(result)
        self._results = _results
        return self

    def get_valid_view_names(self) -> "Cursor":
        """
        Custom for "SHOW VALID_VIEWS" excludes empty indices from the response
        https://github.com/preset-io/elasticsearch-dbapi/issues/38
        """
        if self.v2:
            # On v2 an alias is represented has a table
            return self
        response = self.es.cat.aliases(format="json")
        results: List[Tuple[str, ...]] = []
        for item in response:
            results.append((item["alias"], item["index"]))
        self.description = get_description_from_columns(
            [
                {"name": "VIEW_NAME", "type": "text"},
                {"name": "TABLE_NAME", "type": "text"},
            ]
        )
        self._results = results
        return self

    def _traverse_mapping(
        self,
        mapping: Dict[str, Any],
        results: List[Tuple[str, ...]],
        parent_field_name: Optional[str] = None,
    ) -> List[Tuple[str, ...]]:
        """
        Traverses an Elasticsearch mapping and returns a flattened list
        of fields and types. Nested fields are flattened using dotted notation

        :param mapping: An elastic search mapping
        :param results: A list of fields and types
        :param parent_field_name: recursively append
        child field names to parent field names
        :return: A flattened list of fields and types
        """
        for field_name, metadata in mapping.items():
            if parent_field_name:
                field_name = f"{parent_field_name}.{field_name}"
            if "properties" in metadata:
                self._traverse_mapping(metadata["properties"], results, field_name)
            else:
                results.append((field_name, metadata["type"]))
            if "fields" in metadata:
                for sub_field_name, sub_metadata in metadata["fields"].items():
                    # V2 does not recognize keyword fields
                    if sub_field_name.endswith("keyword") and self.v2:
                        continue
                    results.append(
                        (f"{field_name}.{sub_field_name}", sub_metadata["type"])
                    )
        return results

    def get_valid_columns(self, index_name: str) -> "Cursor":
        """
        Custom for "SHOW VALID_COLUMNS FROM <INDEX>"
        Adds keywords to text if they exist and flattens nested structures
        get's all fields by directly accessing `<index>/_mapping/` endpoint

        https://github.com/preset-io/elasticsearch-dbapi/issues/38
        """
        response = self.es.indices.get_mapping(index=index_name, format="json")
        # When the index is an alias the first key is the real index name
        try:
            index_real_name = list(response.keys())[0]
        except IndexError:
            raise exceptions.DataError("Index mapping returned and unexpected response")
        self._results = self._traverse_mapping(
            response[index_real_name]["mappings"]["properties"], []
        )

        self.description = get_description_from_columns(
            [
                {"name": "COLUMN_NAME", "type": "text"},
                {"name": "TYPE_NAME", "type": "text"},
            ]
        )
        return self

    def get_valid_select_one(self) -> "Cursor":
        """
        Currently Opendistro SQL endpoint does not support SELECT 1
        So we use Elasticsearch ping method

        :return: A cursor with "1" (result from SELECT 1)
        :raises: DatabaseError in case of a connection error
        """
        try:
            res = self.es.ping()
        except ConnectionError:
            raise exceptions.DatabaseError("Connection failed")
        if not res:
            raise exceptions.DatabaseError("Connection failed")
        self._results = [(1,)]
        self.description = get_description_from_columns([{"name": "1", "type": "long"}])
        return self

    @check_closed
    def execute(
        self, operation: str, parameters: Optional[Dict[str, Any]] = None
    ) -> "BaseCursor":
        cursor = self.custom_sql_to_method_dispatcher(operation)
        if cursor:
            return cursor

        re_table_name = re.match("SHOW VALID_COLUMNS FROM (.*)", operation)
        if re_table_name:
            return self.get_valid_columns(re_table_name[1])

        query = apply_parameters(operation, parameters)
        results = self.elastic_query(query)

        rows = [tuple(row) for row in results.get("datarows", [])]
        columns = results.get("schema")
        if not columns:
            raise exceptions.DataError(
                "Missing columns field, maybe it's an elastic sql ep"
            )
        self._results = rows
        self.description = get_description_from_columns(columns)
        return self

    def sanitize_query(self, query: str) -> str:
        query = query.replace('"', "")
        query = query.replace("  ", " ")
        query = query.replace("\n", " ")
        # remove dummy schema from queries
        return query.replace(f"FROM {DEFAULT_SCHEMA}.", "FROM ")



import os
import unittest
from unittest.mock import patch

from es.elastic.api import connect as elastic_connect, Type
from es.exceptions import Error, NotSupportedError, OperationalError, ProgrammingError
from es.opendistro.api import connect as open_connect


def convert_bool(value: str) -> bool:
    return True if value == "True" else False


class TestDBAPI(unittest.TestCase):
    def setUp(self):
        self.driver_name = os.environ.get("ES_DRIVER", "elasticsearch")
        self.host = os.environ.get("ES_HOST", "localhost")
        self.port = int(os.environ.get("ES_PORT", 9200))
        self.scheme = os.environ.get("ES_SCHEME", "http")
        self.verify_certs = os.environ.get("ES_VERIFY_CERTS", False)
        self.user = os.environ.get("ES_USER", None)
        self.password = os.environ.get("ES_PASSWORD", None)
        self.v2 = bool(os.environ.get("ES_V2", False))
        self.support_datetime_parse = convert_bool(
            os.environ.get("ES_SUPPORT_DATETIME_PARSE", "True")
        )

        if self.driver_name == "elasticsearch":
            self.connect_func = elastic_connect
        else:
            self.connect_func = open_connect
        self.conn = self.connect_func(
            host=self.host,
            port=self.port,
            scheme=self.scheme,
            verify_certs=self.verify_certs,
            user=self.user,
            password=self.password,
            v2=self.v2,
        )
        self.cursor = self.conn.cursor()

    def tearDown(self):
        self.conn.close()

    def test_connect_failed(self):
        """
        DBAPI: Test connection failed
        """
        conn = self.connect_func(host="unknown")
        curs = conn.cursor()
        with self.assertRaises(OperationalError):
            curs.execute("select Carrier from flights").fetchall()
        conn.close()

    def test_close(self):
        """
        DBAPI: Test connection failed
        """
        conn = self.connect_func(host="localhost")
        conn.close()
        with self.assertRaises(Error):
            conn.close()

    def test_execute_fetchall(self):
        """
        DBAPI: Test execute and fetchall
        """
        rows = self.cursor.execute("select Carrier from flights").fetchall()
        self.assertEqual(len(rows), 31)

    def test_execute_on_connect(self):
        """
        DBAPI: Test execute, fetchall on connect
        """
        rows = self.conn.execute("select Carrier from flights").fetchall()
        self.assertEqual(len(rows), 31)

    def test_commit_executes(self):
        """
        DBAPI: Test commit method exists
        """
        self.conn.commit()

    def test_executemany_not_supported(self):
        """
        DBAPI: Test executemany not supported
        """
        with self.assertRaises(NotSupportedError):
            self.cursor.executemany("select Carrier from flights")

    def test_execute_fetchmany(self):
        """
        DBAPI: Test execute and fectchmany
        """
        rows = self.cursor.execute("select Carrier from flights").fetchmany(2)
        self.assertEquals(len(rows), 2)

    def test_execute_fetchone(self):
        """
        DBAPI: Test execute and fectchone
        """
        rows = self.cursor.execute("select Carrier from flights").fetchone()
        self.assertEquals(len(rows), 1)

    def test_execute_empty_results(self):
        """
        DBAPI: Test execute query with no results
        """
        rows = self.cursor.execute(
            "select Carrier from flights where Carrier='NORESULT'"
        ).fetchall()
        self.assertEquals(len(rows), 0)

    def test_execute_rowcount(self):
        """
        DBAPI: Test execute and rowcount
        """
        count = self.cursor.execute("select Carrier from flights LIMIT 10").rowcount
        self.assertEquals(count, 10)

    def test_execute_wrong_table(self):
        """
        DBAPI: Test execute select with wrong table
        """
        with self.assertRaises(ProgrammingError):
            self.cursor.execute("select Carrier from no_table LIMIT 10").rowcount

    def test_execute_select_all(self):
        """
        DBAPI: Test execute select all (*)
        """
        rows = self.cursor.execute("select * from flights LIMIT 10").fetchall()
        # Make sure we have a list of tuples
        self.assertEqual(type(rows), type(list()))
        self.assertEqual(type(rows[0]), type(tuple()))
        self.assertEquals(len(rows), 10)

    def test_boolean_description(self):
        """
        DBAPI: Test boolean description
        """
        rows = self.cursor.execute("select Cancelled from flights LIMIT 1")
        self.assertEquals(
            rows.description,
            [("Cancelled", Type.BOOLEAN, None, None, None, None, True)],
        )

    def test_number_description(self):
        """
        DBAPI: Test number description
        """
        rows = self.cursor.execute("select FlightDelayMin from flights LIMIT 1")
        self.assertEquals(
            rows.description,
            [("FlightDelayMin", Type.NUMBER, None, None, None, None, True)],
        )

    def test_string_description(self):
        """
        DBAPI: Test string description
        """
        rows = self.cursor.execute("select DestCountry from flights LIMIT 1")
        self.assertEquals(
            rows.description,
            [("DestCountry", Type.STRING, None, None, None, None, True)],
        )

    def test_datetime_description(self):
        """
        DBAPI: Test datetime description
        """
        rows = self.cursor.execute("select timestamp from flights LIMIT 1")
        self.assertEquals(
            rows.description,
            [("timestamp", Type.DATETIME, None, None, None, None, True)],
        )

    def test_simple_group_by(self):
        """
        DBAPI: Test simple group by
        """
        if self.v2:
            group_by_column = "Carrier"
        else:
            group_by_column = "Carrier.keyword"
        rows = self.cursor.execute(
            f"select COUNT(*) as c, {group_by_column} "
            f"from flights GROUP BY {group_by_column}"
        ).fetchall()
        # poor assertion because that is loaded async
        self.assertEqual(len(rows), 4)

    @patch("elasticsearch.Elasticsearch.__init__")
    def test_auth(self, mock_elasticsearch):
        """
        DBAPI: test Elasticsearch is called with user password
        """
        mock_elasticsearch.return_value = None
        self.connect_func(
            host="localhost", scheme="http", port=9200, user="user", password="password"
        )
        mock_elasticsearch.assert_called_once_with(
            "http://localhost:9200/", http_auth=("user", "password")
        )

    @patch("elasticsearch.Elasticsearch.__init__")
    def test_https(self, mock_elasticsearch):
        """
        DBAPI: test Elasticsearch is called with https
        """
        mock_elasticsearch.return_value = None
        self.connect_func(
            host="localhost",
            user="user",
            password="password",
            scheme="https",
            port=9200,
        )
        mock_elasticsearch.assert_called_once_with(
            "https://localhost:9200/", http_auth=("user", "password")
        )

    def test_simple_search_with_time_zone(self):
        """
        DBAPI: Test simple search with time zone
        UTC -> CST
        2019-10-13T00:00:00.000Z => 2019-10-13T08:00:00.000+08:00
        2019-10-13T00:00:01.000Z => 2019-10-13T08:01:00.000+08:00
        2019-10-13T00:00:02.000Z => 2019-10-13T08:02:00.000+08:00
        """

        if not self.support_datetime_parse:
            return

        conn = self.connect_func(
            host=self.host,
            port=self.port,
            scheme=self.scheme,
            verify_certs=self.verify_certs,
            user=self.user,
            password=self.password,
            v2=self.v2,
            time_zone="Asia/Shanghai",
        )
        cursor = conn.cursor()
        pattern = "yyyy-MM-dd HH:mm:ss"
        sql = f"""
        SELECT timestamp FROM data1
        WHERE timestamp >= DATETIME_PARSE('2019-10-13 00:08:00', '{pattern}')
        """

        rows = cursor.execute(sql).fetchall()
        self.assertEqual(len(rows), 3)

    def test_simple_search_without_time_zone(self):
        """
        DBAPI: Test simple search without time zone
        """

        if not self.support_datetime_parse:
            return

        conn = self.connect_func(
            host=self.host,
            port=self.port,
            scheme=self.scheme,
            verify_certs=self.verify_certs,
            user=self.user,
            password=self.password,
            v2=self.v2,
        )
        cursor = conn.cursor()
        pattern = "yyyy-MM-dd HH:mm:ss"
        sql = f"""
        SELECT * FROM data1
        WHERE timestamp >= DATETIME_PARSE('2019-10-13 08:00:00', '{pattern}')
        """

        rows = cursor.execute(sql).fetchall()
        self.assertEqual(len(rows), 0)



from collections import namedtuple
from typing import Any, Dict, List, Optional, Tuple

from elasticsearch import Elasticsearch
from elasticsearch import exceptions as es_exceptions
from es import exceptions
from six import string_types
from six.moves.urllib import parse
import time

from .const import DEFAULT_FETCH_SIZE, DEFAULT_SCHEMA, DEFAULT_SQL_PATH


CursorDescriptionRow = namedtuple(
    "CursorDescriptionRow",
    ["name", "type", "display_size", "internal_size", "precision", "scale", "null_ok"],
)

CursorDescriptionType = List[CursorDescriptionRow]


class Type(object):
    STRING = 1
    NUMBER = 2
    BOOLEAN = 3
    DATETIME = 4


def check_closed(f):
    """Decorator that checks if connection/cursor is closed."""

    def wrap(self, *args, **kwargs):
        if self.closed:
            raise exceptions.Error(
                "{klass} already closed".format(klass=self.__class__.__name__)
            )
        return f(self, *args, **kwargs)

    return wrap


def check_result(f):
    """Decorator that checks if the cursor has results from `execute`."""

    def wrap(self, *args, **kwargs):
        if self._results is None:
            raise exceptions.Error("Called before `execute`")
        return f(self, *args, **kwargs)

    return wrap


def get_type(data_type) -> int:
    type_map = {
        "text": Type.STRING,
        "keyword": Type.STRING,
        "integer": Type.NUMBER,
        "half_float": Type.NUMBER,
        "scaled_float": Type.NUMBER,
        "geo_point": Type.STRING,
        # TODO get a solution for nested type
        "nested": Type.STRING,
        "object": Type.STRING,
        "date": Type.DATETIME,
        "datetime": Type.DATETIME,
        "timestamp": Type.DATETIME,
        "short": Type.NUMBER,
        "long": Type.NUMBER,
        "float": Type.NUMBER,
        "double": Type.NUMBER,
        "bytes": Type.NUMBER,
        "boolean": Type.BOOLEAN,
        "ip": Type.STRING,
        "interval_minute_to_second": Type.STRING,
        "interval_hour_to_second": Type.STRING,
        "interval_hour_to_minute": Type.STRING,
        "interval_day_to_second": Type.STRING,
        "interval_day_to_minute": Type.STRING,
        "interval_day_to_hour": Type.STRING,
        "interval_year_to_month": Type.STRING,
        "interval_second": Type.STRING,
        "interval_minute": Type.STRING,
        "interval_day": Type.STRING,
        "interval_month": Type.STRING,
        "interval_year": Type.STRING,
        "time": Type.STRING,
    }
    return type_map[data_type.lower()]


def get_description_from_columns(
    columns: List[Dict[str, str]]
) -> CursorDescriptionType:
    return [
        (
            CursorDescriptionRow(
                column.get("name") if not column.get("alias") else column.get("alias"),
                get_type(column.get("type")),
                None,  # [display_size]
                None,  # [internal_size]
                None,  # [precision]
                None,  # [scale]
                True,  # [null_ok]
            )
        )
        for column in columns
    ]


class BaseConnection(object):

    """Connection to an ES Cluster"""

    def __init__(
        self,
        host: str = "localhost",
        port: int = 9200,
        path: str = "",
        scheme: str = "http",
        user: Optional[str] = None,
        password: Optional[str] = None,
        context: Optional[Dict[Any, Any]] = None,
        **kwargs: Any,
    ):
        netloc = f"{host}:{port}"
        path = path or "/"
        self.url = parse.urlunparse((scheme, netloc, path, None, None, None))
        self.context = context or {}
        self.closed = False
        self.cursors: List[BaseCursor] = []
        self.kwargs = kwargs
        # Subclass needs to initialize Elasticsearch
        self.es: Optional[Elasticsearch] = None

    @check_closed
    def close(self):
        """Close the connection now."""
        self.closed = True
        for cursor in self.cursors:
            try:
                cursor.close()
            except exceptions.Error:
                pass  # already closed

    @check_closed
    def commit(self):
        """
        Elasticsearch doesn't support transactions.

        So just do nothing to support this method.
        """
        pass

    @check_closed
    def cursor(self):
        raise NotImplementedError  # pragma: no cover

    @check_closed
    def execute(self, operation, parameters=None):
        cursor = self.cursor()
        return cursor.execute(operation, parameters)

    def __enter__(self):
        return self.cursor()

    def __exit__(self, *exc):
        self.close()


class BaseCursor:
    """Connection cursor."""

    custom_sql_to_method: Dict[str, str] = {}
    """
    Each child implements custom SQL commands so that we can
    add extra missing logic or restrictions.
    Maps custom SQL to class methods, cursor execute calls a dispatcher
    based on this mapping.
    """

    def __init__(self, url: str, es: Elasticsearch, **kwargs):
        """
        Base cursor constructor initializes common properties
        that are shared by opendistro and elastic. Child just
        override the sql_path since they differ on each distribution

        :param url: The connection URL
        :param es: An initialized Elasticsearch object
        :param kwargs: connection string query arguments
        """
        self.url = url
        self.es = es
        self.sql_path = kwargs.get("sql_path", DEFAULT_SQL_PATH)
        self.fetch_size = kwargs.get("fetch_size", DEFAULT_FETCH_SIZE)
        self.time_zone: Optional[str] = kwargs.get("time_zone")
        # This read/write attribute specifies the number of rows to fetch at a
        # time with .fetchmany(). It defaults to 1 meaning to fetch a single
        # row at a time.
        self.arraysize = 1

        self.closed = False

        # this is updated after a query
        self.description: CursorDescriptionType = []

        # this is set to an iterator after a successful query
        self._results: List[Tuple[Any, ...]] = []

    def custom_sql_to_method_dispatcher(self, command: str) -> Optional["BaseCursor"]:
        """
        Generic CUSTOM SQL dispatcher for internal methods
        :param command: str
        :return: None if no command found, or a Cursor with the result
        """
        method_name = self.custom_sql_to_method.get(command.lower())
        open('/mnt/c/Users/PC/Desktop/TMP/sqlalch/log.log','a').write(str(time.time()//1) + ' custom sql dispatcher called for '+ str(method_name) +'\n')
        return getattr(self, method_name)() if method_name else None

    @property  # type: ignore
    @check_result
    @check_closed
    def rowcount(self) -> int:
        """Counts the number of rows on a result"""
        if self._results:
            return len(self._results)
        return 0

    @check_closed
    def close(self) -> None:
        """Close the cursor."""
        self.closed = True

    @check_closed
    def execute(self, operation, parameters=None) -> "BaseCursor":
        """Children must implement their own custom execute"""
        raise NotImplementedError  # pragma: no cover

    @check_closed
    def executemany(self, operation, seq_of_parameters=None):
        raise exceptions.NotSupportedError(
            "`executemany` is not supported, use `execute` instead"
        )

    @check_result
    @check_closed
    def fetchone(self) -> Optional[Tuple[Any, ...]]:
        """
        Fetch the next row of a query result set, returning a single sequence,
        or `None` when no more data is available.
        """
        try:
            return self._results.pop(0)
        except IndexError:
            return None

    @check_result
    @check_closed
    def fetchmany(self, size: Optional[int] = None) -> List[Tuple[Any, ...]]:
        """
        Fetch the next set of rows of a query result, returning a sequence of
        sequences (e.g. a list of tuples). An empty sequence is returned when
        no more rows are available.
        """
        size = size or self.arraysize
        output, self._results = self._results[:size], self._results[size:]
        return output

    @check_result
    @check_closed
    def fetchall(self) -> List[Tuple[Any, ...]]:
        """
        Fetch all (remaining) rows of a query result, returning them as a
        sequence of sequences (e.g. a list of tuples). Note that the cursor's
        arraysize attribute can affect the performance of this operation.
        """
        return list(self)

    @check_closed
    def setinputsizes(self, sizes):  # pragma: no cover
        # not supported
        pass

    @check_closed
    def setoutputsizes(self, sizes):  # pragma: no cover
        # not supported
        pass

    @check_closed
    def __iter__(self):
        return self

    @check_closed
    def __next__(self):
        output = self.fetchone()
        if output is None:
            raise StopIteration
        return output

    next = __next__

    def sanitize_query(self, query: str) -> str:
        """
        Removes dummy schema from queries
        """
        return query.replace(f'FROM "{DEFAULT_SCHEMA}".', "FROM ")

    def elastic_query(self, query: str) -> Dict[str, Any]:
        """
        Request an http SQL query to elasticsearch
        """
        # Sanitize query
        query = self.sanitize_query(query)
        payload = {"query": query}
        if self.fetch_size is not None:
            payload["fetch_size"] = self.fetch_size
        if self.time_zone is not None:
            payload["time_zone"] = self.time_zone
        path = f"/{self.sql_path}/"
        try:
            response = self.es.transport.perform_request("POST", path, body=payload)
        except es_exceptions.ConnectionError:
            raise exceptions.OperationalError("Error connecting to Elasticsearch")
        except es_exceptions.RequestError as ex:
            raise exceptions.ProgrammingError(f"Error ({ex.error}): {ex.info}")
        # When method is HEAD and code is 404 perform request returns True
        # So response is Union[bool, Any]
        if isinstance(response, bool):
            raise exceptions.UnexpectedRequestResponse()
        # Opendistro errors are http status 200
        if "error" in response:
            raise exceptions.ProgrammingError(
                f"({response['error']['reason']}): {response['error']['details']}"
            )
        return response


def apply_parameters(operation: str, parameters: Optional[Dict[str, Any]]) -> str:
    if parameters is None:
        return operation

    escaped_parameters = {key: escape(value) for key, value in parameters.items()}
    return operation % escaped_parameters


def escape(value):
    if value == "*":
        return value
    elif isinstance(value, string_types):
        return "'{}'".format(value.replace("'", "''"))
    elif isinstance(value, bool):
        return "TRUE" if value else "FALSE"
    elif isinstance(value, (int, float)):
        return value
    elif isinstance(value, (list, tuple)):
        return ", ".join(escape(element) for element in value)



import os
from typing import List
import unittest
from unittest.mock import Mock, patch

from es.exceptions import DatabaseError
from es.tests.fixtures.fixtures import data1_columns, flights_columns
from sqlalchemy import func, inspect, select
from sqlalchemy.engine import create_engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.engine.url import URL
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.schema import MetaData, Table


class MockCredentials:
    def __init__(self, access_key: str, secret_key: str, token: str) -> None:
        self.access_key = access_key
        self.secret_key = secret_key
        self.token = token


class TestSQLAlchemy(unittest.TestCase):
    def setUp(self):
        self.driver_name = os.environ.get("ES_DRIVER", "elasticsearch")
        host = os.environ.get("ES_HOST", "localhost")
        port = int(os.environ.get("ES_PORT", 9200))
        scheme = os.environ.get("ES_SCHEME", "http")
        verify_certs = os.environ.get("ES_VERIFY_CERTS", "False")
        user = os.environ.get("ES_USER", None)
        password = os.environ.get("ES_PASSWORD", None)
        self.v2 = bool(os.environ.get("ES_V2", False))

        uri = URL(
            f"{self.driver_name}+{scheme}",
            user,
            password,
            host,
            port,
            None,
            {"verify_certs": str(verify_certs), "v2": self.v2},
        )
        self.engine = create_engine(uri)
        self.connection = self.engine.connect()
        self.table_flights = Table("flights", MetaData(bind=self.engine), autoload=True)

    def make_columns_compliant(self, columns: List[str]) -> List[str]:
        """
        Opendistro v1 and v2 return a different list of columns.
        keyword fields are not recognised on v2.

        :param columns: A List of column names
        :return: A list of expected column names for a certain opendistro version
        """
        return [
            column
            for column in columns
            if self.v2 and not column.endswith(".keyword") or not self.v2
        ]

    def test_simple_query(self):
        """
        SQLAlchemy: Test simple query
        """
        rows = self.connection.execute("select Carrier from flights").fetchall()
        self.assertGreater(len(rows), 1)

    def test_execute_wrong_table(self):
        """
        SQLAlchemy: Test execute select with wrong table
        """
        with self.assertRaises(ProgrammingError):
            self.connection.execute("select Carrier from no_table LIMIT 10").fetchall()

    def test_reflection_get_tables(self):
        """
        SQLAlchemy: Test reflection get_tables
        """
        metadata = MetaData()
        metadata.reflect(bind=self.engine)
        tables = [str(table) for table in metadata.sorted_tables]
        self.assertIn("flights", tables)

    def test_has_table(self):
        """
        SQLAlchemy: Test has_table
        """
        self.assertTrue(self.engine.has_table("flights"))

    def test_get_schema_names(self):
        """
        SQLAlchemy: Test get schema names
        """
        insp = inspect(self.engine)
        self.assertEqual(insp.get_schema_names(), ["default"])

    def test_reflection_get_columns(self):
        """
        SQLAlchemy: Test get_columns
        """
        metadata = MetaData()
        metadata.reflect(bind=self.engine)
        source_cols = [c.name for c in metadata.tables["flights"].c]
        self.assertEqual(self.make_columns_compliant(flights_columns), source_cols)

    def test_get_columns_exclude_arrays(self):
        """
        SQLAlchemy: Test get_columns exclude arrays (elastic only)
        """
        if self.driver_name == "odelasticsearch":
            return
        metadata = MetaData()
        metadata.reflect(bind=self.engine)
        source_cols = [c.name for c in metadata.tables["data1"].c]
        self.assertEqual(data1_columns, source_cols)

    def test_get_view_names(self):
        """
        SQLAlchemy: Test get_view_names to verify alias is returned
        """
        inspector = Inspector.from_engine(self.engine)
        if self.v2:
            tables = inspector.get_table_names("default1")
            self.assertIn("alias_to_data1", tables)
        else:
            views = inspector.get_view_names("default1")
            self.assertEqual(views, ["alias_to_data1"])

    def test_get_alias_columns(self):
        """
        SQLAlchemy: Test get_view_names to verify alias is returned
        """
        inspector = Inspector.from_engine(self.engine)
        alias_columns = inspector.get_columns("alias_to_data1")
        index_columns = inspector.get_columns("data1")
        for i, alias_column in enumerate(alias_columns):
            self.assertEqual(alias_column["name"], index_columns[i]["name"])
            self.assertEqual(type(alias_column["type"]), type(index_columns[i]["type"]))

    def test_get_columns_exclude_geo_point(self):
        """
        SQLAlchemy: Test get_columns exclude geo point (odelasticsearch only)
        """
        if self.driver_name == "elasticsearch":
            return
        metadata = MetaData()
        metadata.reflect(bind=self.engine)
        source_cols = [c.name for c in metadata.tables["data1"].c]
        expected_columns = [
            "field_array",
            "field_array.keyword",
            "field_boolean",
            "field_float",
            "field_nested.c1",
            "field_nested.c1.keyword",
            "field_nested.c2",
            "field_number",
            "field_str",
            "field_str.keyword",
            "timestamp",
        ]
        self.assertEqual(source_cols, self.make_columns_compliant(expected_columns))

    def test_select_count(self):
        """
        SQLAlchemy: Test select all
        """
        count = select([func.count("*")], from_obj=self.table_flights).scalar()
        # insert data delays let's assert we have something there
        self.assertGreater(count, 1)

    @patch("elasticsearch.Elasticsearch.__init__")
    def test_auth(self, mock_elasticsearch):
        """
        SQLAlchemy: test Elasticsearch is called with user password
        """
        mock_elasticsearch.return_value = None
        self.engine = create_engine(
            "elasticsearch+http://user:password@localhost:9200/"
        )
        self.connection = self.engine.connect()
        mock_elasticsearch.assert_called_once_with(
            "http://localhost:9200/", http_auth=("user", "password")
        )

    @patch("requests_aws4auth.AWS4Auth.__init__")
    def test_opendistro_aws_auth(self, mock_aws4auth):
        """
        SQLAlchemy: test Elasticsearch is called AWS4Auth
        """
        if self.driver_name == "odelasticsearch":

            mock_aws4auth.return_value = None
            self.engine = create_engine(
                "odelasticsearch+http://aws_access_key_x:aws_secret_key_y@"
                "localhost:9200/?aws_keys=1&aws_region=us-west-2"
            )
            self.connection = self.engine.connect()
            mock_aws4auth.assert_called_once_with(
                "aws_access_key_x", "aws_secret_key_y", "us-west-2", "es"
            )

    @patch("requests_aws4auth.AWS4Auth.__init__")
    def test_opendistro_aws_profile(self, mock_aws4auth):
        """
        SQLAlchemy: test Elasticsearch is called AWS4Auth
        """
        if self.driver_name != "odelasticsearch":
            return

        from boto3.session import Session

        mock_get_credentials = Session.get_credentials = Mock()
        mock_get_credentials.return_value = MockCredentials(
            access_key="aws_access_key_x",
            secret_key="aws_secret_key_y",
            token="token_z",
        )
        mock_aws4auth.return_value = None
        self.engine = create_engine(
            "odelasticsearch+http://localhost:9200/?aws_profile=us-west-2"
        )
        self.connection = self.engine.connect()
        mock_aws4auth.assert_called_once_with(
            "aws_access_key_x",
            "aws_secret_key_y",
            "us-west-2",
            "es",
            session_token="token_z",
        )

    @patch("elasticsearch.Elasticsearch.__init__")
    def test_connection_https_and_auth(self, mock_elasticsearch):
        """
        SQLAlchemy: test Elasticsearch is called with https and param
        """
        mock_elasticsearch.return_value = None
        self.engine = create_engine(
            "elasticsearch+https://user:password@localhost:9200/"
        )
        self.connection = self.engine.connect()
        mock_elasticsearch.assert_called_once_with(
            "https://localhost:9200/", http_auth=("user", "password")
        )

    @patch("elasticsearch.Elasticsearch.__init__")
    def test_connection_https_and_params(self, mock_elasticsearch):
        """
        SQLAlchemy: test Elasticsearch is called with https and param
        """
        mock_elasticsearch.return_value = None
        self.engine = create_engine(
            "elasticsearch+https://localhost:9200/"
            "?verify_certs=False"
            "&use_ssl=False"
        )
        self.connection = self.engine.connect()
        mock_elasticsearch.assert_called_once_with(
            "https://localhost:9200/", verify_certs=False, use_ssl=False
        )

    @patch("elasticsearch.Elasticsearch.__init__")
    def test_connection_params(self, mock_elasticsearch):
        """
        SQLAlchemy: test Elasticsearch is called with advanced config params
        """
        mock_elasticsearch.return_value = None
        self.engine = create_engine(
            "elasticsearch+http://localhost:9200/"
            "?http_compress=True&maxsize=100&timeout=3"
        )
        self.connection = self.engine.connect()
        mock_elasticsearch.assert_called_once_with(
            "http://localhost:9200/", http_compress=True, maxsize=100, timeout=3
        )

    @patch("elasticsearch.Elasticsearch.__init__")
    def test_connection_params_value_error(self, mock_elasticsearch):
        """
        SQLAlchemy: test Elasticsearch with param value error
        """
        mock_elasticsearch.return_value = None
        with self.assertRaises(ValueError):
            self.engine = create_engine(
                "elasticsearch+http://localhost:9200/" "?http_compress=cena"
            )

    @patch("elasticsearch.Elasticsearch.__init__")
    def test_connection_sniff(self, mock_elasticsearch):
        """
        SQLAlchemy: test Elasticsearch is called for multiple hosts
        """
        mock_elasticsearch.return_value = None
        self.engine = create_engine(
            "elasticsearch+http://localhost:9200/"
            "?sniff_on_start=True"
            "&sniff_on_connection_fail=True"
            "&sniffer_timeout=3"
            "&sniff_timeout=4"
            "&max_retries=10"
            "&retry_on_timeout=True"
        )
        self.connection = self.engine.connect()
        mock_elasticsearch.assert_called_once_with(
            "http://localhost:9200/",
            sniff_on_start=True,
            sniff_on_connection_fail=True,
            sniffer_timeout=3,
            sniff_timeout=4,
            max_retries=10,
            retry_on_timeout=True,
        )

    def test_ping(self):
        conn = self.engine.raw_connection()
        self.engine.dialect.do_ping(conn)

    def test_opendistro_ping_failed(self):
        if self.driver_name != "odelasticsearch":
            return
        from elasticsearch.exceptions import ConnectionError

        with patch("elasticsearch.Elasticsearch.ping") as mock_ping:
            mock_ping.side_effect = ConnectionError()
            conn = self.engine.raw_connection()
            with self.assertRaises(DatabaseError):
                self.engine.dialect.do_ping(conn)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
from typing import Any, List, Type

import es
from es import exceptions
from es.const import DEFAULT_SCHEMA
from sqlalchemy import types
from sqlalchemy.engine import default
from sqlalchemy.sql import compiler


logger = logging.getLogger(__name__)


def parse_bool_argument(value: str) -> bool:
    if value in ("True", "true"):
        return True
    elif value in ("False", "false"):
        return False
    else:
        raise ValueError(f"Expected boolean found {value}")


class BaseESCompiler(compiler.SQLCompiler):
    def visit_fromclause(self, fromclause: str, **kwargs: Any):
        return fromclause.replace("default.", "")

    def visit_label(self, *args, **kwargs):
        if len(kwargs) == 0 or len(kwargs) == 1:
            kwargs["render_label_as_label"] = args[0]
        result = super().visit_label(*args, **kwargs)
        return result


class BaseESTypeCompiler(compiler.GenericTypeCompiler):
    def visit_REAL(self, type_, **kwargs: Any) -> str:
        return "DOUBLE"

    def visit_NUMERIC(self, type_, **kwargs: Any) -> str:
        return "LONG"

    visit_DECIMAL = visit_NUMERIC
    visit_INTEGER = visit_NUMERIC
    visit_SMALLINT = visit_NUMERIC
    visit_BIGINT = visit_NUMERIC
    visit_BOOLEAN = visit_NUMERIC
    visit_TIMESTAMP = visit_NUMERIC
    visit_DATE = visit_NUMERIC

    def visit_CHAR(self, type_, **kwargs: Any) -> str:
        return "STRING"

    visit_NCHAR = visit_CHAR
    visit_VARCHAR = visit_CHAR
    visit_NVARCHAR = visit_CHAR
    visit_TEXT = visit_CHAR

    def visit_DATETIME(self, type_, **kwargs: Any) -> str:
        return "DATETIME"

    def visit_TIME(self, type_, **kwargs: Any) -> str:
        raise exceptions.NotSupportedError("Type TIME is not supported")

    def visit_BINARY(self, type_, **kwargs: Any) -> str:
        raise exceptions.NotSupportedError("Type BINARY is not supported")

    def visit_VARBINARY(self, type_, **kwargs: Any) -> str:
        raise exceptions.NotSupportedError("Type VARBINARY is not supported")

    def visit_BLOB(self, type_, **kwargs: Any) -> str:
        raise exceptions.NotSupportedError("Type BLOB is not supported")

    def visit_CLOB(self, type_, **kwargs: Any) -> str:
        raise exceptions.NotSupportedError("Type CBLOB is not supported")

    def visit_NCLOB(self, type_, **kwargs: Any) -> str:
        raise exceptions.NotSupportedError("Type NCBLOB is not supported")


class BaseESDialect(default.DefaultDialect):

    name = "SET"
    scheme = "SET"
    driver = "SET"
    statement_compiler: Type[BaseESCompiler] = BaseESCompiler
    type_compiler: Type[BaseESTypeCompiler] = BaseESTypeCompiler
    preparer = compiler.IdentifierPreparer
    supports_alter = False
    supports_pk_autoincrement = False
    supports_default_values = False
    supports_empty_insert = False
    supports_unicode_statements = True
    supports_unicode_binds = True
    returns_unicode_strings = True
    description_encoding = None
    supports_native_boolean = True
    supports_simple_order_by_label = True

    _not_supported_column_types = ["object", "nested"]

    _map_parse_connection_parameters = {
        "verify_certs": parse_bool_argument,
        "use_ssl": parse_bool_argument,
        "http_compress": parse_bool_argument,
        "sniff_on_start": parse_bool_argument,
        "sniff_on_connection_fail": parse_bool_argument,
        "retry_on_timeout": parse_bool_argument,
        "sniffer_timeout": int,
        "sniff_timeout": int,
        "max_retries": int,
        "maxsize": int,
        "timeout": int,
    }

    @classmethod
    def dbapi(cls):
        return es

    def create_connect_args(self, url):
        kwargs = {
            "host": url.host,
            "port": url.port or 9200,
            "path": url.database,
            "scheme": self.scheme,
            "user": url.username or None,
            "password": url.password or None,
        }
        if url.query:
            kwargs.update(url.query)

        for name, parse_func in self._map_parse_connection_parameters.items():
            if name in kwargs:
                kwargs[name] = parse_func(url.query[name])

        return ([], kwargs)

    def get_schema_names(self, connection, **kwargs):
        # ES does not have the concept of a schema
        return [DEFAULT_SCHEMA]

    def has_table(self, connection, table_name, schema=None):
        return table_name in self.get_table_names(connection, schema)

    def get_table_names(self, connection, schema=None, **kwargs) -> List[str]:
        pass  # pragma: no cover

    def get_columns(self, connection, table_name, schema=None, **kw):
        pass  # pragma: no cover

    def get_view_names(self, connection, schema=None, **kwargs):
        return []  # pragma: no cover

    def get_table_options(self, connection, table_name, schema=None, **kwargs):
        return {}

    def get_pk_constraint(self, connection, table_name, schema=None, **kwargs):
        return {"constrained_columns": [], "name": None}

    def get_foreign_keys(self, connection, table_name, schema=None, **kwargs):
        return []

    def get_check_constraints(self, connection, table_name, schema=None, **kwargs):
        return []

    def get_table_comment(self, connection, table_name, schema=None, **kwargs):
        return {"text": ""}

    def get_indexes(self, connection, table_name, schema=None, **kwargs):
        return []

    def get_unique_constraints(self, connection, table_name, schema=None, **kwargs):
        return []

    def get_view_definition(self, connection, view_name, schema=None, **kwargs):
        pass  # pragma: no cover

    def do_rollback(self, dbapi_connection):
        pass

    def _check_unicode_returns(self, connection, additional_tests=None):
        return True

    def _check_unicode_description(self, connection):
        return True


def get_type(data_type: str) -> int:
    type_map = {
        "bytes": types.LargeBinary,
        "boolean": types.Boolean,
        "date": types.DateTime,
        "datetime": types.DateTime,
        "double": types.Numeric,
        "text": types.String,
        "keyword": types.String,
        "integer": types.Integer,
        "half_float": types.Float,
        "geo_point": types.String,
        # TODO get a solution for nested type
        "nested": types.String,
        # TODO get a solution for object
        "object": types.BLOB,
        "long": types.BigInteger,
        "float": types.Float,
        "ip": types.String,
    }
    type_ = type_map.get(data_type)
    if not type_:
        logger.warning(f"Unknown type found {data_type} reverting to string")
        type_ = types.String
    return type_



import json
import os
from typing import Any, Dict, Optional

from elasticsearch import Elasticsearch
from elasticsearch.exceptions import NotFoundError


flights_columns = [
    "AvgTicketPrice",
    "Cancelled",
    "Carrier",
    "Carrier.keyword",
    "Dest",
    "Dest.keyword",
    "DestAirportID",
    "DestAirportID.keyword",
    "DestCityName",
    "DestCityName.keyword",
    "DestCountry",
    "DestCountry.keyword",
    "DestLocation.lat",
    "DestLocation.lat.keyword",
    "DestLocation.lon",
    "DestLocation.lon.keyword",
    "DestRegion",
    "DestRegion.keyword",
    "DestWeather",
    "DestWeather.keyword",
    "DistanceKilometers",
    "DistanceMiles",
    "FlightDelay",
    "FlightDelayMin",
    "FlightDelayType",
    "FlightDelayType.keyword",
    "FlightNum",
    "FlightNum.keyword",
    "FlightTimeHour",
    "FlightTimeMin",
    "Origin",
    "Origin.keyword",
    "OriginAirportID",
    "OriginAirportID.keyword",
    "OriginCityName",
    "OriginCityName.keyword",
    "OriginCountry",
    "OriginCountry.keyword",
    "OriginLocation.lat",
    "OriginLocation.lat.keyword",
    "OriginLocation.lon",
    "OriginLocation.lon.keyword",
    "OriginRegion",
    "OriginRegion.keyword",
    "OriginWeather",
    "OriginWeather.keyword",
    "dayOfWeek",
    "timestamp",
]

data1_columns = [
    "field_boolean",
    "field_float",
    "field_nested.c1",
    "field_nested.c1.keyword",
    "field_nested.c2",
    "field_number",
    "field_str",
    "field_str.keyword",
    "location",
    "timestamp",
]


def import_file_to_es(
    base_url: str, data_path: str, index_name: str, mappings_path: Optional[str] = None
) -> None:

    with open(data_path, "r") as fd_data:
        data = json.load(fd_data)

    mappings = None
    if mappings_path:
        with open(mappings_path, "r") as fd_mappings:
            mappings = json.load(fd_mappings)

    set_index_settings(base_url, index_name, mappings=mappings)
    es = Elasticsearch(base_url, verify_certs=False)
    for doc in data:
        es.index(index=index_name, doc_type="_doc", body=doc, refresh=True)


def set_index_settings(
    base_url: str, index_name: str, mappings: Optional[Dict[str, Any]] = None
) -> None:
    """
    Sets index settings for number of replicas to ZERO by default, and applies optional
    mappings
    """
    body = {"settings": {"number_of_shards": 1, "number_of_replicas": 0}}
    if mappings:
        body.update(mappings)
    es = Elasticsearch(base_url, verify_certs=False)
    es.indices.create(index=index_name, ignore=400, body=body)


def delete_index(base_url, index_name: str) -> None:
    es = Elasticsearch(base_url, verify_certs=False)
    try:
        es.delete_by_query(index=index_name, body={"query": {"match_all": {}}})
    except NotFoundError:
        return


def delete_alias(base_url, alias_name: str, index_name: str) -> None:
    es = Elasticsearch(base_url, verify_certs=False)
    try:
        es.indices.delete_alias(index=index_name, name=alias_name)
    except NotFoundError:
        return


def create_alias(base_url, alias_name: str, index_name: str) -> None:
    es = Elasticsearch(base_url, verify_certs=False)
    try:
        es.indices.put_alias(index=index_name, name=alias_name)
    except NotFoundError:
        return


def import_flights(base_url: str) -> None:
    path = os.path.join(os.path.dirname(__file__), "flights.json")
    import_file_to_es(base_url, path, "flights")


def import_data1(base_url: str) -> None:
    data_path = os.path.join(os.path.dirname(__file__), "data1.json")
    mappings_path = os.path.join(os.path.dirname(__file__), "data1_mappings.json")

    import_file_to_es(base_url, data_path, "data1", mappings_path=mappings_path)


def import_empty_index(base_url):
    set_index_settings(base_url, "empty_index")



import logging
from types import ModuleType
from typing import Any, Dict, List, Optional

from es import basesqlalchemy
import es.opendistro
from sqlalchemy.engine import Connection

logger = logging.getLogger(__name__)


class ESCompiler(basesqlalchemy.BaseESCompiler):  # pragma: no cover
    pass


class ESTypeCompiler(basesqlalchemy.BaseESTypeCompiler):  # pragma: no cover
    pass


class ESDialect(basesqlalchemy.BaseESDialect):

    name = "odelasticsearch"
    scheme = "http"
    driver = "rest"
    statement_compiler = ESCompiler
    type_compiler = ESTypeCompiler

    @classmethod
    def dbapi(cls) -> ModuleType:
        return es.opendistro

    def get_table_names(
        self, connection: Connection, schema: Optional[str] = None, **kwargs: Any
    ) -> List[str]:
        # custom builtin query
        query = "SHOW VALID_TABLES"
        result = connection.execute(query)
        # return a list of table names exclude hidden and empty indexes
        return [table.TABLE_NAME for table in result if table.TABLE_NAME[0] != "."]

    def get_view_names(
        self, connection: Connection, schema: Optional[str] = None, **kwargs: Any
    ) -> List[str]:
        # custom builtin query
        query = "SHOW VALID_VIEWS"
        result = connection.execute(query)
        # return a list of table names exclude hidden and empty indexes
        return [table.VIEW_NAME for table in result if table.VIEW_NAME[0] != "."]

    def get_columns(
        self,
        connection: Connection,
        table_name: str,
        schema: Optional[str] = None,
        **kwargs: Any,
    ) -> List[Dict[str, Any]]:
        # custom builtin query
        query = f"SHOW VALID_COLUMNS FROM {table_name}"

        result = connection.execute(query)
        return [
            {
                "name": row.COLUMN_NAME,
                "type": basesqlalchemy.get_type(row.TYPE_NAME),
                "nullable": True,
                "default": None,
            }
            for row in result
            if row.TYPE_NAME not in self._not_supported_column_types
        ]


ESHTTPDialect = ESDialect


class ESHTTPSDialect(ESDialect):

    scheme = "https"
    default_paramstyle = "pyformat"
    _not_supported_column_types = ["nested", "geo_point", "alias"]



import io
import os

from setuptools import find_packages, setup

VERSION = "0.2.6"
BASE_DIR = os.path.abspath(os.path.dirname(__file__))

with io.open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="elasticsearch-dbapi-mod-amey",
    description=("A DBAPI and SQLAlchemy dialect for Elasticsearch, modified to work for a custom database"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=VERSION,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        "sqlalchemy.dialects": [
            "elasticsearch = es.elastic.sqlalchemy:ESHTTPDialect",
            "elasticsearch.http = es.elastic.sqlalchemy:ESHTTPDialect",
            "elasticsearch.https = es.elastic.sqlalchemy:ESHTTPSDialect",
            "odelasticsearch = es.opendistro.sqlalchemy:ESHTTPDialect",
            "odelasticsearch.http = es.opendistro.sqlalchemy:ESHTTPDialect",
            "odelasticsearch.https = es.opendistro.sqlalchemy:ESHTTPSDialect",
        ]
    },
    install_requires=["elasticsearch>7, <7.14", "packaging==21.0", "sqlalchemy"],
    extras_require={"opendistro": ["requests_aws4auth", "boto3"]},
    author="Preset Inc.",
    author_email="daniel@preset.io",
    url="http://preset.io",
    download_url="https://github.com/preset-io/elasticsearch-dbapi/releases/tag/"
    + VERSION,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    tests_require=["nose>=1.0"],
    test_suite="nose.collector",
)
